import binascii

from pymacaroons.binders.base_binder import BaseBinder
from pymacaroons.utils import hmac_concat, truncate_or_pad


class HashSignaturesBinder(BaseBinder):

    def __init__(self, root, key=None):
        super(HashSignaturesBinder, self).__init__(root)
        self.key = key or truncate_or_pad(b'\0')

    def bind_signature(self, signature):
        return hmac_concat(
            self.key,
            binascii.unhexlify(self.root.signature_bytes),
            binascii.unhexlify(signature)
        )



from .caveat import Caveat
from .macaroon import Macaroon
from .verifier import Verifier


__all__ = [
    'Macaroon',
    'Caveat',
    'Verifier',
]


__author__ = 'Evan Cordell'

__version__ = "0.9.2"
__version_info__ = tuple(__version__.split('.'))
__short_version__ = __version__



from base64 import standard_b64encode

from pymacaroons.utils import convert_to_string, convert_to_bytes


class Caveat(object):

    def __init__(self,
                 caveat_id=None,
                 verification_key_id=None,
                 location=None):
        self.caveat_id = caveat_id
        self.verification_key_id = verification_key_id
        self.location = location

    @property
    def caveat_id(self):
        return convert_to_string(self._caveat_id)

    @property
    def verification_key_id(self):
        return self._verification_key_id

    @property
    def location(self):
        return convert_to_string(self._location)

    @caveat_id.setter
    def caveat_id(self, value):
        self._caveat_id = convert_to_bytes(value)

    @verification_key_id.setter
    def verification_key_id(self, value):
        self._verification_key_id = convert_to_bytes(value)

    @location.setter
    def location(self, value):
        self._location = convert_to_bytes(value)

    def first_party(self):
        return self._verification_key_id is None

    def third_party(self):
        return self._verification_key_id is not None

    def to_dict(self):
        return {
            'cid': self.caveat_id,
            'vid': (
                standard_b64encode(self.verification_key_id)
                if self.verification_key_id else None
            ),
            'cl': self.location
        }



from abc import ABCMeta, abstractmethod


class BaseFirstPartyCaveatDelegate(object):
    __metaclass__ = ABCMeta

    def __init__(self, *args, **kwargs):
        super(BaseFirstPartyCaveatDelegate, self).__init__(*args, **kwargs)

    @abstractmethod
    def add_first_party_caveat(self, macaroon, predicate, **kwargs):
        pass


class BaseFirstPartyCaveatVerifierDelegate(object):
    __metaclass__ = ABCMeta

    def __init__(self, *args, **kwargs):
        super(BaseFirstPartyCaveatVerifierDelegate, self).__init__(
            *args, **kwargs
        )

    @abstractmethod
    def verify_first_party_caveat(self, verifier, caveat, signature):
        pass

    @abstractmethod
    def update_signature(self, signature, caveat):
        pass



from base64 import standard_b64encode, standard_b64decode

from libnacl.secret import SecretBox

from pymacaroons.field_encryptors.base_field_encryptor import (
    BaseFieldEncryptor
)
from pymacaroons.utils import (
    truncate_or_pad, convert_to_bytes, convert_to_string
)


class SecretBoxEncryptor(BaseFieldEncryptor):

    def __init__(self, signifier=None, nonce=None):
        super(SecretBoxEncryptor, self).__init__(
            signifier=signifier or 'sbe::'
        )
        self.nonce = nonce

    def encrypt(self, signature, field_data):
        encrypt_key = truncate_or_pad(signature)
        box = SecretBox(key=encrypt_key)
        encrypted = box.encrypt(convert_to_bytes(field_data), nonce=self.nonce)
        return self._signifier + standard_b64encode(encrypted)

    def decrypt(self, signature, field_data):
        key = truncate_or_pad(signature)
        box = SecretBox(key=key)
        encoded = convert_to_bytes(field_data[len(self.signifier):])
        decrypted = box.decrypt(standard_b64decode(encoded))
        return convert_to_string(decrypted)



from abc import ABCMeta, abstractmethod


class BaseThirdPartyCaveatDelegate(object):
    __metaclass__ = ABCMeta

    def __init__(self, *args, **kwargs):
        super(BaseThirdPartyCaveatDelegate, self).__init__(*args, **kwargs)

    @abstractmethod
    def add_third_party_caveat(self,
                               macaroon,
                               location,
                               key,
                               key_id,
                               **kwargs):
        pass


class BaseThirdPartyCaveatVerifierDelegate(object):
    __metaclass__ = ABCMeta

    def __init__(self, *args, **kwargs):
        super(BaseThirdPartyCaveatVerifierDelegate, self).__init__(
            *args, **kwargs
        )

    @abstractmethod
    def verify_third_party_caveat(self,
                                  verifier,
                                  caveat,
                                  root_macaroon,
                                  discharge_macaroons,
                                  signature):
        pass

    @abstractmethod
    def update_signature(self, signature, caveat):
        pass



import binascii

try:
    # py2.7.7+ and py3.3+ have native comparison support
    from hmac import compare_digest as constant_time_compare
except ImportError:
    from pymacaroons.utils import equals as constant_time_compare

from pymacaroons.binders import HashSignaturesBinder
from pymacaroons.exceptions import MacaroonInvalidSignatureException
from pymacaroons.caveat_delegates import (
    FirstPartyCaveatVerifierDelegate,
    ThirdPartyCaveatVerifierDelegate,
)
from pymacaroons.utils import (
    convert_to_bytes,
    convert_to_string,
    generate_derived_key,
    hmac_digest
)


class Verifier(object):

    def __init__(self):
        self.predicates = []
        self.callbacks = [self.verify_exact]
        self.calculated_signature = None
        self.first_party_caveat_verifier_delegate = (
            FirstPartyCaveatVerifierDelegate()
        )
        self.third_party_caveat_verifier_delegate = (
            ThirdPartyCaveatVerifierDelegate()
        )

    def satisfy_exact(self, predicate):
        if predicate is None:
            raise TypeError('Predicate cannot be none.')
        self.predicates.append(convert_to_string(predicate))

    def satisfy_general(self, func):
        if not hasattr(func, '__call__'):
            raise TypeError('General caveat verifiers must be callable.')
        self.callbacks.append(func)

    def verify_exact(self, predicate):
        return predicate in self.predicates

    def verify(self, macaroon, key, discharge_macaroons=None):
        key = generate_derived_key(convert_to_bytes(key))
        return self.verify_discharge(
            macaroon,
            macaroon,
            key,
            discharge_macaroons
        )

    def verify_discharge(self, root, discharge, key, discharge_macaroons=None):
        calculated_signature = hmac_digest(
            key, discharge._identifier
        )

        calculated_signature = self._verify_caveats(
            discharge, discharge_macaroons, calculated_signature
        )

        if root != discharge:
            calculated_signature = binascii.unhexlify(
                HashSignaturesBinder(root).bind_signature(
                    binascii.hexlify(calculated_signature)
                )
            )

        if not self._signatures_match(
                discharge.signature,
                binascii.hexlify(calculated_signature)):
            raise MacaroonInvalidSignatureException('Signatures do not match.')

        return True

    def _verify_caveats(self, macaroon, discharge_macaroons, signature):
        for caveat in macaroon.caveats:
            if self._caveat_met(caveat,
                                macaroon,
                                discharge_macaroons,
                                signature):
                signature = self._update_signature(caveat, signature)
        return signature

    def _caveat_met(self, caveat, macaroon, discharge_macaroons, signature):
        if caveat.first_party():
            return (
                self
                .first_party_caveat_verifier_delegate
                .verify_first_party_caveat(self, caveat, signature)
            )
        else:
            return (
                self
                .third_party_caveat_verifier_delegate
                .verify_third_party_caveat(
                    self, caveat, macaroon, discharge_macaroons, signature
                )
            )

    def _update_signature(self, caveat, signature):
        if caveat.first_party():
            return (
                self
                .first_party_caveat_verifier_delegate
                .update_signature(signature, caveat)
            )
        else:
            return (
                self
                .third_party_caveat_verifier_delegate
                .update_signature(signature, caveat)
            )

    def _signatures_match(self, macaroon_signature, computed_signature):
        return constant_time_compare(
            convert_to_string(macaroon_signature),
            convert_to_string(computed_signature)
        )



from __future__ import unicode_literals

import binascii
import struct
from base64 import urlsafe_b64encode, urlsafe_b64decode

from pymacaroons.utils import convert_to_bytes
from pymacaroons.serializers.base_serializer import BaseSerializer
from pymacaroons.exceptions import MacaroonSerializationException


class BinarySerializer(BaseSerializer):

    PACKET_PREFIX_LENGTH = 4

    def serialize(self, macaroon):
        combined = self._packetize(b'location', macaroon.location)
        combined += self._packetize(b'identifier', macaroon.identifier)

        for caveat in macaroon.caveats:
            combined += self._packetize(b'cid', caveat._caveat_id)

            if caveat._verification_key_id and caveat._location:
                combined += self._packetize(
                    b'vid', caveat._verification_key_id)
                combined += self._packetize(b'cl', caveat._location)

        combined += self._packetize(
            b'signature',
            binascii.unhexlify(macaroon.signature_bytes)
        )
        return urlsafe_b64encode(combined).decode('ascii').rstrip('=')

    def deserialize(self, serialized):
        from pymacaroons.macaroon import Macaroon
        from pymacaroons.caveat import Caveat
        from pymacaroons.exceptions import MacaroonDeserializationException

        macaroon = Macaroon()

        decoded = urlsafe_b64decode(convert_to_bytes(
            serialized + "=" * (-len(serialized) % 4)
        ))

        index = 0

        while index < len(decoded):
            packet_length = int(
                struct.unpack(
                    b"4s",
                    decoded[index:index + self.PACKET_PREFIX_LENGTH]
                )[0],
                16
            )
            packet = decoded[
                index + self.PACKET_PREFIX_LENGTH:index + packet_length
            ]

            key, value = self._depacketize(packet)

            if key == b'location':
                macaroon.location = value
            elif key == b'identifier':
                macaroon.identifier = value
            elif key == b'cid':
                macaroon.caveats.append(Caveat(caveat_id=value))
            elif key == b'vid':
                macaroon.caveats[-1].verification_key_id = value
            elif key == b'cl':
                macaroon.caveats[-1].location = value
            elif key == b'signature':
                macaroon.signature = binascii.hexlify(value)
            else:
                raise MacaroonDeserializationException(
                    'Key {key} not valid key for this format. '
                    'Value: {value}'.format(
                        key=key, value=value
                    )
                )

            index = index + packet_length

        return macaroon

    def _packetize(self, key, data):
        # The 2 covers the space and the newline
        packet_size = self.PACKET_PREFIX_LENGTH + 2 + len(key) + len(data)
        # Ignore the first two chars, 0x
        packet_size_hex = hex(packet_size)[2:]

        if packet_size > 65535:
            raise MacaroonSerializationException(
                'Packet too long for serialization. '
                'Max length is 0xFFFF (65535). '
                'Packet length: 0x{hex_length} ({length}) '
                'Key: {key}'.format(
                    key=key,
                    hex_length=packet_size_hex,
                    length=packet_size
                )
            )

        header = packet_size_hex.zfill(4).encode('ascii')
        packet_content = key + b' ' + convert_to_bytes(data) + b'\n'
        packet = struct.pack(
            convert_to_bytes("4s%ds" % len(packet_content)),
            header,
            packet_content
        )
        return packet

    def _depacketize(self, packet):
        key = packet.split(b' ')[0]
        value = packet[len(key) + 1:-1]
        return (key, value)
